@@ -28,6 +28,7 @@
 #include <enumeration.h>
 #include <qmlprivategate.h>
 
+#include <QDebug>
 #include <QEvent>
 #include <QQmlContext>
 #include <QQmlError>
@@ -43,6 +44,7 @@
 #include <QQmlParserStatus>
 #include <QTextDocument>
 #include <QLibraryInfo>
+#include <QJSValue>
 
 static bool isSimpleExpression(const QString &expression)
 {
@@ -125,7 +127,8 @@ void ObjectNodeInstance::initializePropertyWatcher(const ObjectNodeInstance::Poi
     m_signalSpy.setObjectNodeInstance(objectNodeInstance);
 }
 
-void ObjectNodeInstance::initialize(const ObjectNodeInstance::Pointer &objectNodeInstance)
+void ObjectNodeInstance::initialize(const ObjectNodeInstance::Pointer &objectNodeInstance,
+                                    InstanceContainer::NodeFlags /*flags*/)
 {
     initializePropertyWatcher(objectNodeInstance);
     QmlPrivateGate::registerNodeInstanceMetaObject(objectNodeInstance->object(), objectNodeInstance->nodeInstanceServer()->engine());
@@ -253,7 +256,7 @@ bool ObjectNodeInstance::isAnchoredByChildren() const
 
 QPair<PropertyName, ServerNodeInstance> ObjectNodeInstance::anchor(const PropertyName &/*name*/) const
 {
-    return qMakePair(PropertyName(), ServerNodeInstance());
+    return {PropertyName(), ServerNodeInstance()};
 }
 
 
@@ -262,11 +265,17 @@ static bool isList(const QQmlProperty &property)
     return property.propertyTypeCategory() == QQmlProperty::List;
 }
 
+static bool isQJSValue(const QQmlProperty &property)
+{
+    return property.isValid() && !strcmp(property.propertyTypeName(), "QJSValue");
+}
+
 static bool isObject(const QQmlProperty &property)
 {
-    return (property.propertyTypeCategory() == QQmlProperty::Object) ||
-            //QVariant can also store QObjects. Lets trust our model.
-           (QLatin1String(property.propertyTypeName()) == QLatin1String("QVariant"));
+    /* QVariant and QJSValue can also store QObjects. Lets trust our model. */
+    return property.isValid() && (property.propertyTypeCategory() == QQmlProperty::Object
+            || !strcmp(property.propertyTypeName(), "QVariant")
+            || isQJSValue(property));
 }
 
 static QVariant objectToVariant(QObject *object)
@@ -335,7 +344,14 @@ void ObjectNodeInstance::addToNewProperty(QObject *object, QObject *newParent, c
 
         list.append(object);
     } else if (isObject(property)) {
-        property.write(objectToVariant(object));
+        if (isQJSValue(property)) /* In this case we have to explcitly generate and convert a QJSValue */
+            property.write(QVariant::fromValue(engine()->newQObject(object)));
+        else
+            property.write(objectToVariant(object));
+
+        if (QQuickItem *item = qobject_cast<QQuickItem *>(object))
+            if (QQuickItem *newParentItem = qobject_cast<QQuickItem *>(newParent))
+                item->setParentItem(newParentItem);
     }
 
     Q_ASSERT(objectToVariant(object).isValid());
@@ -390,7 +406,7 @@ QVariant ObjectNodeInstance::convertEnumToValue(const QVariant &value, const Pro
         QQmlExpression expression(context(), object(), enumeration.toString());
         adjustedValue =  expression.evaluate();
         if (expression.hasError())
-            qDebug() << "Enumeration can not be evaluated:" << object() << name << enumeration;
+            qDebug() << "Enumeration cannot be evaluated:" << object() << name << enumeration;
     }
     return adjustedValue;
 }
@@ -416,7 +432,7 @@ void ObjectNodeInstance::setPropertyVariant(const PropertyName &name, const QVar
     if (oldValue.type() == QVariant::Url) {
         QUrl url = oldValue.toUrl();
         QString path = url.toLocalFile();
-        if (QFileInfo(path).exists() && nodeInstanceServer() && !path.isEmpty())
+        if (QFileInfo::exists(path) && nodeInstanceServer() && !path.isEmpty())
             nodeInstanceServer()->removeFilePropertyFromFileSystemWatcher(object(), name, path);
     }
 
@@ -433,7 +449,7 @@ void ObjectNodeInstance::setPropertyVariant(const PropertyName &name, const QVar
     if (newValue.type() == QVariant::Url) {
         QUrl url = newValue.toUrl();
         QString path = url.toLocalFile();
-        if (QFileInfo(path).exists() && nodeInstanceServer() && !path.isEmpty())
+        if (QFileInfo::exists(path) && nodeInstanceServer() && !path.isEmpty())
             nodeInstanceServer()->addFilePropertyToFileSystemWatcher(object(), name, path);
     }
 }
@@ -601,7 +617,52 @@ ObjectNodeInstance::Pointer ObjectNodeInstance::create(QObject *object)
 
 QObject *ObjectNodeInstance::createPrimitive(const QString &typeName, int majorNumber, int minorNumber, QQmlContext *context)
 {
-    return QmlPrivateGate::createPrimitive(typeName, majorNumber, minorNumber, context);
+    QString polishTypeName = typeName;
+    if (typeName == "QtQuick.Controls/Popup"
+            || typeName == "QtQuick.Controls/Drawer"
+            || typeName == "QtQuick.Controls/Dialog"
+            || typeName == "QtQuick.Controls/Menu"
+            || typeName == "QtQuick.Controls/Pane"
+            || typeName == "QtQuick.Controls/ToolTip")
+        polishTypeName = "QtQuick/Item";
+
+    const QHash<QString, QString> mockHash = {{"QtQuick.Controls/SwipeView","qrc:/qtquickplugin/mockfiles/SwipeView.qml"},
+                                             {"QtQuick.Dialogs/Dialog","qrc:/qtquickplugin/mockfiles/Dialog.qml"}};
+
+    QObject *object = nullptr;
+
+    if (mockHash.contains(typeName))
+        object = QmlPrivateGate::createComponent(mockHash.value(typeName), context);
+    else
+        object = QmlPrivateGate::createPrimitive(polishTypeName, majorNumber, minorNumber, context);
+
+    /* Let's try to create the primitive from source, since with incomplete meta info this might be a pure
+     * QML type. This is the case for example if a C++ type is mocked up with a QML file.
+     */
+
+    if (!object)
+        object = createPrimitiveFromSource(polishTypeName, majorNumber, minorNumber, context);
+
+    return object;
+}
+
+QObject *ObjectNodeInstance::createPrimitiveFromSource(const QString &typeName, int majorNumber, int minorNumber, QQmlContext *context)
+{
+    if (typeName.isEmpty())
+        return 0;
+
+    QStringList parts = typeName.split("/");
+    const QString unqualifiedTypeName = parts.last();
+    parts.removeLast();
+
+    if (parts.isEmpty())
+        return 0;
+
+    QString importString = parts.join(".") + " " + QString::number(majorNumber) + "." + QString::number(minorNumber);
+    if (importString == "QtQuick 1.0") /* Workaround for implicit QQml import */
+        importString = "QtQuick 2.0";
+    QString source = "import " + importString + "\n" + unqualifiedTypeName + " {\n" + "}\n";
+    return createCustomParserObject(source, "", context);
 }
 
 QObject *ObjectNodeInstance::createComponentWrap(const QString &nodeSource, const QByteArray &importCode, QQmlContext *context)
@@ -638,16 +699,16 @@ static inline QString fixComponentPathForIncompatibleQt(const QString &component
     if (componentPath.contains(importString)) {
         int index = componentPath.indexOf(importString) + 8;
         const QString relativeImportPath = componentPath.right(componentPath.length() - index);
-        QString fixedComponentPath = QLibraryInfo::location(QLibraryInfo::ImportsPath) + relativeImportPath;
+        QString fixedComponentPath = QLibraryInfo::location(QLibraryInfo::Qml2ImportsPath) + relativeImportPath;
         fixedComponentPath.replace(QLatin1Char('\\'), QLatin1Char('/'));
-        if (QFileInfo(fixedComponentPath).exists())
+        if (QFileInfo::exists(fixedComponentPath))
             return fixedComponentPath;
         QString fixedPath = QFileInfo(fixedComponentPath).path();
         if (fixedPath.endsWith(QLatin1String(".1.0"))) {
         //plugin directories might contain the version number
             fixedPath.chop(4);
             fixedPath += QLatin1Char('/') + QFileInfo(componentPath).fileName();
-            if (QFileInfo(fixedPath).exists())
+            if (QFileInfo::exists(fixedPath))
                 return fixedPath;
         }
     }
